@@ -17,12 +17,13 @@
 -------------------------------------------------------------
 
 module Data.IntTrie 
-    ( IntTrie, identity, apply, modify, overwrite )
+    ( IntTrie, identity, apply, modify, modify', overwrite )
 where
 
 import Control.Applicative
 import Data.Bits
 import Data.Function (fix)
+import Data.Monoid (Monoid(..))
 
 -- | A trie from integers to values of type a. 
 -- 
@@ -38,6 +39,10 @@ instance Applicative BitTrie where
     pure x = fix (\g -> BitTrie x g g)
     ~(BitTrie f fl fr) <*> ~(BitTrie x xl xr) = BitTrie (f x) (fl <*> xl) (fr <*> xr)
 
+instance Monoid a => Monoid (BitTrie a) where
+    mempty = pure mempty
+    mappend = liftA2 mappend
+
 instance Functor IntTrie where
     fmap f ~(IntTrie neg z pos) = IntTrie (fmap f neg) (f z) (fmap f pos)
 
@@ -46,16 +51,20 @@ instance Applicative IntTrie where
     IntTrie fneg fz fpos <*> IntTrie xneg xz xpos = 
         IntTrie (fneg <*> xneg) (fz xz) (fpos <*> xpos)
 
+instance Monoid a => Monoid (IntTrie a) where
+    mempty = pure mempty
+    mappend = liftA2 mappend
+
 -- | Apply the trie to an argument.  This is the semantic map.
 apply :: (Ord b, Num b, Bits b) => IntTrie a -> b -> a
-apply ~(IntTrie neg z pos) x =
+apply (IntTrie neg z pos) x =
     case compare x 0 of
         LT -> applyPositive neg (-x)
         EQ -> z
         GT -> applyPositive pos x
 
 applyPositive :: (Num b, Bits b) => BitTrie a -> b -> a
-applyPositive ~(BitTrie one even odd) x
+applyPositive (BitTrie one even odd) x
     | x == 1 = one
     | testBit x 0 = applyPositive odd (x `shiftR` 1)
     | otherwise   = applyPositive even (x `shiftR` 1)
@@ -88,6 +97,22 @@ modifyPositive x f ~(BitTrie one even odd)
     | testBit x 0 = BitTrie one even (modifyPositive (x `shiftR` 1) f odd)
     | otherwise   = BitTrie one (modifyPositive (x `shiftR` 1) f even) odd
 
+
+-- | Modify the function at one point (strict version)
+modify' :: (Ord b, Num b, Bits b) => b -> (a -> a) -> IntTrie a -> IntTrie a
+modify' x f (IntTrie neg z pos) =
+    case compare x 0 of
+        LT -> (IntTrie $! modifyPositive' (-x) f neg) z pos
+        EQ -> (IntTrie neg $! f z) pos
+        GT -> IntTrie neg z $! modifyPositive' x f pos
+
+modifyPositive' :: (Num b, Bits b) => b -> (a -> a) -> BitTrie a -> BitTrie a
+modifyPositive' x f (BitTrie one even odd)
+    | x == 1      = (BitTrie $! f one) even odd
+    | testBit x 0 = BitTrie one even $! modifyPositive' (x `shiftR` 1) f odd
+    | otherwise   = (BitTrie one $! modifyPositive' (x `shiftR` 1) f even) odd
+
+
 -- | Overwrite the function at one point
 --
 -- > overwrite i x = modify i (const x)
